"""
ritm_annotation base module.

This is the principal module of the ritm_annotation project.
here you put your main classes and objects.

Be creative! do whatever you want!

If you want to replace this with a Flask application run:

    $ make init

and then choose `flask` as template.
"""

# example constant variable
NAME = "ritm_annotation"
